@@ -21,8 +21,7 @@ from cinderclient import utils
 
 _quota_resources = ['volumes', 'snapshots', 'gigabytes',
                     'backups', 'backup_gigabytes',
-                    'consistencygroups', 'per_volume_gigabytes',
-                    'groups', ]
+                    'per_volume_gigabytes', 'groups', ]
 _quota_infos = ['Type', 'In_use', 'Reserved', 'Limit']
 
 
@@ -85,9 +84,10 @@ def find_consistencygroup(cs, consistencygroup):
     return utils.find_resource(cs.consistencygroups, consistencygroup)
 
 
-def find_group(cs, group):
+def find_group(cs, group, **kwargs):
     """Gets a group by name or ID."""
-    return utils.find_resource(cs.groups, group)
+    kwargs['is_group'] = True
+    return utils.find_resource(cs.groups, group, **kwargs)
 
 
 def find_cgsnapshot(cs, cgsnapshot):
@@ -150,7 +150,9 @@ def extract_filters(args):
             (key, value) = f.split('=', 1)
             if value.startswith('{') and value.endswith('}'):
                 value = _build_internal_dict(value[1:-1])
-        filters[key] = value
+            filters[key] = value
+        else:
+            print("WARNING: Ignoring the filter %s while showing result." % f)
 
     return filters
 
