@@ -28,7 +28,7 @@ REQUEST_ID = 'req-test-request-id'
 def _stub_volume(*args, **kwargs):
     volume = {
         "migration_status": None,
-        "attachments": [{'server_id': 1234}],
+        "attachments": [{u'server_id': u'1234'}],
         "links": [
             {
                 "href": "http://localhost/v2/fake/volumes/1234",
@@ -369,7 +369,7 @@ class FakeHTTPClient(base_client.HTTPClient):
         # add fake request-id header
         headers['x-openstack-request-id'] = REQUEST_ID
         if self.version_header:
-            headers['OpenStack-API-version'] = version_header
+            headers['OpenStack-API-version'] = self.version_header
         r = utils.TestResponse({
             "status_code": status,
             "text": body,
@@ -535,6 +535,8 @@ class FakeHTTPClient(base_client.HTTPClient):
         elif action == 'os-volume_upload_image':
             assert 'image_name' in body[action]
             _body = body
+        elif action == 'revert':
+            assert 'snapshot_id' in body[action]
         else:
             raise AssertionError("Unexpected action: %s" % action)
         return (resp, {}, _body)
@@ -625,7 +627,6 @@ class FakeHTTPClient(base_client.HTTPClient):
                           'gigabytes': 1,
                           'backups': 1,
                           'backup_gigabytes': 1,
-                          'consistencygroups': 1,
                           'per_volume_gigabytes': 1, }})
 
     def get_os_quota_sets_test_defaults(self):
@@ -637,7 +638,6 @@ class FakeHTTPClient(base_client.HTTPClient):
                           'gigabytes': 1,
                           'backups': 1,
                           'backup_gigabytes': 1,
-                          'consistencygroups': 1,
                           'per_volume_gigabytes': 1, }})
 
     def put_os_quota_sets_test(self, body, **kw):
@@ -652,7 +652,6 @@ class FakeHTTPClient(base_client.HTTPClient):
                           'gigabytes': 1,
                           'backups': 1,
                           'backup_gigabytes': 1,
-                          'consistencygroups': 2,
                           'per_volume_gigabytes': 1, }})
 
     def delete_os_quota_sets_1234(self, **kw):
@@ -668,13 +667,11 @@ class FakeHTTPClient(base_client.HTTPClient):
     def get_os_quota_class_sets_test(self, **kw):
         return (200, {}, {'quota_class_set': {
                           'class_name': 'test',
-                          'metadata_items': [],
                           'volumes': 1,
                           'snapshots': 1,
                           'gigabytes': 1,
                           'backups': 1,
                           'backup_gigabytes': 1,
-                          'consistencygroups': 1,
                           'per_volume_gigabytes': 1, }})
 
     def put_os_quota_class_sets_test(self, body, **kw):
@@ -683,13 +680,11 @@ class FakeHTTPClient(base_client.HTTPClient):
                               required=['class_name'])
         return (200, {}, {'quota_class_set': {
                           'class_name': 'test',
-                          'metadata_items': [],
                           'volumes': 2,
                           'snapshots': 2,
                           'gigabytes': 1,
                           'backups': 1,
                           'backup_gigabytes': 1,
-                          'consistencygroups': 2,
                           'per_volume_gigabytes': 1}})
 
     #
@@ -770,6 +765,13 @@ class FakeHTTPClient(base_client.HTTPClient):
     def put_types_1(self, **kw):
         return self.get_types_1()
 
+    def put_types_3(self, **kw):
+        return (200, {}, {'volume_type': {'id': 3,
+                          'name': 'test-type-2',
+                          'description': 'test_type-3-desc',
+                          'is_public': True,
+                          'extra_specs': {}}})
+
     #
     # VolumeAccess
     #
@@ -1072,9 +1074,9 @@ class FakeHTTPClient(base_client.HTTPClient):
             },
         ]
         if host:
-            services = filter(lambda i: i['host'] == host, services)
+            services = [i for i in services if i['host'] == host]
         if binary:
-            services = filter(lambda i: i['binary'] == binary, services)
+            services = [i for i in services if i['binary'] == binary]
         return (200, {}, {'services': services})
 
     def put_os_services_enable(self, body, **kw):
