@@ -1,3 +1,4 @@
+# -*- coding: utf-8 -*-
 # Copyright (c) 2013 OpenStack Foundation
 # All Rights Reserved.
 #
@@ -13,15 +14,42 @@
 #    License for the specific language governing permissions and limitations
 #    under the License.
 
+
+# NOTE(geguileo): For v3 we cannot mock any of the following methods
+#   - utils.find_volume
+#   - shell_utils.find_backup
+#   - shell_utils.find_volume_snapshot
+#   - shell_utils.find_group
+#   - shell_utils.find_group_snapshot
+# because we are caching them in cinderclient.v3.shell:RESET_STATE_RESOURCES
+# which means that our tests could fail depending on the mocking and loading
+# order.
+#
+# Alternatives are:
+#   - Mock utils.find_resource when we have only 1 call to that method
+#   - Use an auxiliary method that will call original method for irrelevant
+#     calls.  Example from test_revert_to_snapshot:
+#         original = client_utils.find_resource
+#
+#         def find_resource(manager, name_or_id, **kwargs):
+#             if isinstance(manager, volume_snapshots.SnapshotManager):
+#                 return volume_snapshots.Snapshot(self,
+#                                                  {'id': '5678',
+#                                                   'volume_id': '1234'})
+#             return original(manager, name_or_id, **kwargs)
+
 import ddt
 import fixtures
 import mock
 from requests_mock.contrib import fixture as requests_mock_fixture
+import six
 
 from cinderclient import client
 from cinderclient import exceptions
 from cinderclient import shell
+from cinderclient import utils as cinderclient_utils
 from cinderclient.v3 import volumes
+from cinderclient.v3 import volume_snapshots
 from cinderclient.tests.unit import utils
 from cinderclient.tests.unit.v3 import fakes
 from cinderclient.tests.unit.fixture_data import keystone_client
@@ -95,6 +123,14 @@ class ShellTest(utils.TestCase):
              'list --filters metadata={key1:value1}',
          'expected':
              '/volumes/detail?metadata=%7B%27key1%27%3A+%27value1%27%7D'},
+        {'command':
+             'list --filters name~=456',
+         'expected':
+             '/volumes/detail?name%7E=456'},
+        {'command':
+             u'list --filters name~=Σ',
+         'expected':
+             '/volumes/detail?name%7E=%CE%A3'},
         # testcases for list group
         {'command':
              'group-list --filters name=456',
@@ -104,6 +140,10 @@ class ShellTest(utils.TestCase):
              'group-list --filters status=available',
          'expected':
              '/groups/detail?status=available'},
+        {'command':
+             'group-list --filters name~=456',
+         'expected':
+             '/groups/detail?name%7E=456'},
         # testcases for list group-snapshot
         {'command':
              'group-snapshot-list --status=error --filters status=available',
@@ -113,6 +153,10 @@ class ShellTest(utils.TestCase):
              'group-snapshot-list --filters availability_zone=123',
          'expected':
              '/group_snapshots/detail?availability_zone=123'},
+        {'command':
+             'group-snapshot-list --filters status~=available',
+         'expected':
+             '/group_snapshots/detail?status%7E=available'},
         # testcases for list message
         {'command':
              'message-list --event_id=123 --filters event_id=456',
@@ -122,6 +166,10 @@ class ShellTest(utils.TestCase):
              'message-list --filters request_id=123',
          'expected':
              '/messages?request_id=123'},
+        {'command':
+             'message-list --filters request_id~=123',
+         'expected':
+             '/messages?request_id%7E=123'},
         # testcases for list attachment
         {'command':
              'attachment-list --volume-id=123 --filters volume_id=456',
@@ -131,6 +179,10 @@ class ShellTest(utils.TestCase):
              'attachment-list --filters mountpoint=123',
          'expected':
              '/attachments?mountpoint=123'},
+        {'command':
+             'attachment-list --filters volume_id~=456',
+         'expected':
+             '/attachments?volume_id%7E=456'},
         # testcases for list backup
         {'command':
              'backup-list --volume-id=123 --filters volume_id=456',
@@ -140,6 +192,10 @@ class ShellTest(utils.TestCase):
              'backup-list --filters name=123',
          'expected':
              '/backups/detail?name=123'},
+        {'command':
+             'backup-list --filters volume_id~=456',
+         'expected':
+             '/backups/detail?volume_id%7E=456'},
         # testcases for list snapshot
         {'command':
              'snapshot-list --volume-id=123 --filters volume_id=456',
@@ -149,6 +205,10 @@ class ShellTest(utils.TestCase):
              'snapshot-list --filters name=123',
          'expected':
              '/snapshots/detail?name=123'},
+        {'command':
+             'snapshot-list --filters volume_id~=456',
+         'expected':
+             '/snapshots/detail?volume_id%7E=456'},
         # testcases for get pools
         {'command':
              'get-pools --filters name=456 --detail',
@@ -213,15 +273,36 @@ class ShellTest(utils.TestCase):
                                       'mountpoint': '/123',
                                       'initiator': 'aabbccdd',
                                       'platform': 'x86_xx'},
+                        'volume_uuid': '1234'}},
+              {'cmd': 'abc 1233',
+               'body': {'instance_uuid': '1233',
+                        'connector': {},
                         'volume_uuid': '1234'}})
+    @mock.patch('cinderclient.utils.find_resource')
     @ddt.unpack
-    def test_attachment_create(self, cmd, body):
+    def test_attachment_create(self, mock_find_volume, cmd, body):
+        mock_find_volume.return_value = volumes.Volume(self,
+                                                       {'id': '1234'},
+                                                       loaded=True)
         command = '--os-volume-api-version 3.27 attachment-create '
         command += cmd
         self.run_command(command)
         expected = {'attachment': body}
+        self.assertTrue(mock_find_volume.called)
         self.assert_called('POST', '/attachments', body=expected)
 
+    @mock.patch.object(volumes.VolumeManager, 'findall')
+    def test_attachment_create_duplicate_name_vol(self, mock_findall):
+        found = [volumes.Volume(self, {'id': '7654', 'name': 'abc'},
+                                loaded=True),
+                 volumes.Volume(self, {'id': '9876', 'name': 'abc'},
+                                loaded=True)]
+        mock_findall.return_value = found
+        self.assertRaises(exceptions.CommandError,
+                          self.run_command,
+                          '--os-volume-api-version 3.27 '
+                          'attachment-create abc 789')
+
     @ddt.data({'cmd': '',
                'expected': ''},
               {'cmd': '--volume-id 1234',
@@ -245,6 +326,24 @@ class ShellTest(utils.TestCase):
         self.run_command(command)
         self.assert_called('GET', '/attachments%s' % expected)
 
+    def test_revert_to_snapshot(self):
+        original = cinderclient_utils.find_resource
+
+        def find_resource(manager, name_or_id, **kwargs):
+            if isinstance(manager, volume_snapshots.SnapshotManager):
+                return volume_snapshots.Snapshot(self,
+                                                 {'id': '5678',
+                                                  'volume_id': '1234'})
+            return original(manager, name_or_id, **kwargs)
+
+        with mock.patch('cinderclient.utils.find_resource',
+                        side_effect=find_resource):
+            self.run_command(
+                '--os-volume-api-version 3.40 revert-to-snapshot 5678')
+
+        self.assert_called('POST', '/volumes/1234/action',
+                           body={'revert': {'snapshot_id': '5678'}})
+
     def test_attachment_show(self):
         self.run_command('--os-volume-api-version 3.27 attachment-show 1234')
         self.assert_called('GET', '/attachments/1234')
@@ -427,6 +526,11 @@ class ShellTest(utils.TestCase):
                          'group-show 1234')
         self.assert_called('GET', '/groups/1234')
 
+    def test_group_show_with_list_volume(self):
+        self.run_command('--os-volume-api-version 3.25 '
+                         'group-show 1234 --list-volume')
+        self.assert_called('GET', '/groups/1234?list_volume=True')
+
     @ddt.data(True, False)
     def test_group_delete(self, delete_vol):
         cmd = '--os-volume-api-version 3.13 group-delete 1234'
@@ -679,7 +783,7 @@ class ShellTest(utils.TestCase):
         self.assert_called_anytime('DELETE', '/messages/1234')
         self.assert_called_anytime('DELETE', '/messages/12345')
 
-    @mock.patch('cinderclient.utils.find_volume')
+    @mock.patch('cinderclient.utils.find_resource')
     def test_delete_metadata(self, mock_find_volume):
         mock_find_volume.return_value = volumes.Volume(self,
                                                        {'id': '1234',
@@ -703,3 +807,112 @@ class ShellTest(utils.TestCase):
             command += ' --withreplication %s' % replication
         self.run_command(command)
         self.assert_called('GET', '/os-services')
+
+    def test_group_enable_replication(self):
+        cmd = '--os-volume-api-version 3.38 group-enable-replication 1234'
+        self.run_command(cmd)
+        expected = {'enable_replication': {}}
+        self.assert_called('POST', '/groups/1234/action', body=expected)
+
+    def test_group_disable_replication(self):
+        cmd = '--os-volume-api-version 3.38 group-disable-replication 1234'
+        self.run_command(cmd)
+        expected = {'disable_replication': {}}
+        self.assert_called('POST', '/groups/1234/action', body=expected)
+
+    @ddt.data((False, None), (True, None),
+              (False, "backend1"), (True, "backend1"),
+              (False, "default"), (True, "default"))
+    @ddt.unpack
+    def test_group_failover_replication(self, attach_vol, backend):
+        attach = '--allow-attached-volume ' if attach_vol else ''
+        backend_id = ('--secondary-backend-id ' + backend) if backend else ''
+        cmd = ('--os-volume-api-version 3.38 group-failover-replication 1234 '
+               + attach + backend_id)
+        self.run_command(cmd)
+        expected = {'failover_replication':
+                    {'allow_attached_volume': attach_vol,
+                     'secondary_backend_id': backend if backend else None}}
+        self.assert_called('POST', '/groups/1234/action', body=expected)
+
+    def test_group_list_replication_targets(self):
+        cmd = ('--os-volume-api-version 3.38 group-list-replication-targets'
+               ' 1234')
+        self.run_command(cmd)
+        expected = {'list_replication_targets': {}}
+        self.assert_called('POST', '/groups/1234/action', body=expected)
+
+    @mock.patch('cinderclient.v3.services.ServiceManager.get_log_levels')
+    def test_service_get_log_before_3_32(self, get_levels_mock):
+        self.assertRaises(SystemExit,
+                          self.run_command, '--os-volume-api-version 3.28 '
+                         'service-get-log')
+        get_levels_mock.assert_not_called()
+
+    @mock.patch('cinderclient.v3.services.ServiceManager.get_log_levels')
+    @mock.patch('cinderclient.utils.print_list')
+    def test_service_get_log_no_params(self, print_mock, get_levels_mock):
+        self.run_command('--os-volume-api-version 3.32 service-get-log')
+        get_levels_mock.assert_called_once_with('', '', '')
+        print_mock.assert_called_once_with(get_levels_mock.return_value,
+                                           ('Binary', 'Host', 'Prefix',
+                                            'Level'))
+
+    @ddt.data('*', 'cinder-api', 'cinder-volume', 'cinder-scheduler',
+              'cinder-backup')
+    @mock.patch('cinderclient.v3.services.ServiceManager.get_log_levels')
+    @mock.patch('cinderclient.utils.print_list')
+    def test_service_get_log(self, binary, print_mock, get_levels_mock):
+        server = 'host1'
+        prefix = 'sqlalchemy'
+
+        self.run_command('--os-volume-api-version 3.32 service-get-log '
+                         '--binary %s --server %s --prefix %s' % (
+                             binary, server, prefix))
+        get_levels_mock.assert_called_once_with(binary, server, prefix)
+        print_mock.assert_called_once_with(get_levels_mock.return_value,
+                                           ('Binary', 'Host', 'Prefix',
+                                            'Level'))
+
+    @mock.patch('cinderclient.v3.services.ServiceManager.set_log_levels')
+    def test_service_set_log_before_3_32(self, set_levels_mock):
+        self.assertRaises(SystemExit,
+                          self.run_command, '--os-volume-api-version 3.28 '
+                         'service-set-log debug')
+        set_levels_mock.assert_not_called()
+
+    @mock.patch('cinderclient.v3.services.ServiceManager.set_log_levels')
+    @mock.patch('cinderclient.shell.CinderClientArgumentParser.error')
+    def test_service_set_log_missing_required(self, error_mock,
+                                              set_levels_mock):
+        error_mock.side_effect = SystemExit
+        self.assertRaises(SystemExit,
+                          self.run_command, '--os-volume-api-version 3.32 '
+                          'service-set-log')
+        set_levels_mock.assert_not_called()
+        # Different error message from argparse library in Python 2 and 3
+        if six.PY3:
+            msg = 'the following arguments are required: <log-level>'
+        else:
+            msg = 'too few arguments'
+        error_mock.assert_called_once_with(msg)
+
+    @ddt.data('debug', 'DEBUG', 'info', 'INFO', 'warning', 'WARNING', 'error',
+              'ERROR')
+    @mock.patch('cinderclient.v3.services.ServiceManager.set_log_levels')
+    def test_service_set_log_min_params(self, level, set_levels_mock):
+        self.run_command('--os-volume-api-version 3.32 '
+                         'service-set-log %s' % level)
+        set_levels_mock.assert_called_once_with(level, '', '', '')
+
+    @ddt.data('*', 'cinder-api', 'cinder-volume', 'cinder-scheduler',
+              'cinder-backup')
+    @mock.patch('cinderclient.v3.services.ServiceManager.set_log_levels')
+    def test_service_set_log_levels(self, binary, set_levels_mock):
+        level = 'debug'
+        server = 'host1'
+        prefix = 'sqlalchemy.'
+        self.run_command('--os-volume-api-version 3.32 '
+                         'service-set-log %s --binary %s --server %s '
+                         '--prefix %s' % (level, binary, server, prefix))
+        set_levels_mock.assert_called_once_with(level, binary, server, prefix)
