@@ -17,11 +17,13 @@
 from __future__ import print_function
 
 import argparse
+import collections
 import os
 
 from oslo_utils import strutils
 import six
 
+import cinderclient
 from cinderclient import api_versions
 from cinderclient import base
 from cinderclient import exceptions
@@ -30,11 +32,154 @@ from cinderclient import utils
 
 from cinderclient.v2.shell import *  # flake8: noqa
 
+FILTER_DEPRECATED = ("This option is deprecated and will be removed in "
+                     "newer release. Please use '--filters' option which "
+                     "is introduced since 3.33 instead.")
+
+
+@api_versions.wraps('3.33')
+@utils.arg('--resource',
+           metavar='<resource>',
+           default=None,
+           help='Show enabled filters for specified resource. Default=None.')
+def do_list_filters(cs, args):
+        """List enabled filters.
+
+        Symbol '~' after filter key means it supports inexact filtering.
+        """
+        filters = cs.resource_filters.list(resource=args.resource)
+        shell_utils.print_resource_filter_list(filters)
+
+
+@utils.arg('--all-tenants',
+           metavar='<all_tenants>',
+           nargs='?',
+           type=int,
+           const=1,
+           default=0,
+           help='Shows details for all tenants. Admin only.')
+@utils.arg('--all_tenants',
+           nargs='?',
+           type=int,
+           const=1,
+           help=argparse.SUPPRESS)
+@utils.arg('--name',
+           metavar='<name>',
+           default=None,
+           help="Filters results by a name. Default=None. "
+                "%s" % FILTER_DEPRECATED)
+@utils.arg('--status',
+           metavar='<status>',
+           default=None,
+           help="Filters results by a status. Default=None. "
+                "%s" % FILTER_DEPRECATED)
+@utils.arg('--volume-id',
+           metavar='<volume-id>',
+           default=None,
+           help="Filters results by a volume ID. Default=None. "
+                "%s" % FILTER_DEPRECATED)
+@utils.arg('--volume_id',
+           help=argparse.SUPPRESS)
+@utils.arg('--marker',
+           metavar='<marker>',
+           default=None,
+           help='Begin returning backups that appear later in the backup '
+                'list than that represented by this id. '
+                'Default=None.')
+@utils.arg('--limit',
+           metavar='<limit>',
+           default=None,
+           help='Maximum number of backups to return. Default=None.')
+@utils.arg('--sort',
+           metavar='<key>[:<direction>]',
+           default=None,
+           help=(('Comma-separated list of sort keys and directions in the '
+                  'form of <key>[:<asc|desc>]. '
+                  'Valid keys: %s. '
+                  'Default=None.') % ', '.join(base.SORT_KEY_VALUES)))
+@utils.arg('--filters',
+           type=six.text_type,
+           nargs='*',
+           start_version='3.33',
+           metavar='<key=value>',
+           default=None,
+           help="Filter key and value pairs. Please use 'cinder list-filters' "
+                "to check enabled filters from server. Use 'key~=value' for "
+                "inexact filtering if the key supports. Default=None.")
+def do_backup_list(cs, args):
+    """Lists all backups."""
+    # pylint: disable=function-redefined
+
+    search_opts = {
+        'all_tenants': args.all_tenants,
+        'name': args.name,
+        'status': args.status,
+        'volume_id': args.volume_id,
+    }
+
+    # Update search option with `filters`
+    if hasattr(args, 'filters') and args.filters is not None:
+        search_opts.update(shell_utils.extract_filters(args.filters))
+
+    backups = cs.backups.list(search_opts=search_opts,
+                              marker=args.marker,
+                              limit=args.limit,
+                              sort=args.sort)
+    shell_utils.translate_volume_snapshot_keys(backups)
+    columns = ['ID', 'Volume ID', 'Status', 'Name', 'Size', 'Object Count',
+               'Container']
+    if args.sort:
+        sortby_index = None
+    else:
+        sortby_index = 0
+    utils.print_list(backups, columns, sortby_index=sortby_index)
+
+
+@utils.arg('--detail',
+           action='store_true',
+           help='Show detailed information about pools.')
+@utils.arg('--filters',
+           type=str,
+           nargs='*',
+           start_version='3.33',
+           metavar='<key=value>',
+           default=None,
+           help="Filter key and value pairs. Please use 'cinder list-filters' "
+                "to check enabled filters from server, Default=None.")
+def do_get_pools(cs, args):
+    """Show pool information for backends. Admin only."""
+    # pylint: disable=function-redefined
+    search_opts = {}
+    # Update search option with `filters`
+    if hasattr(args, 'filters') and args.filters is not None:
+        search_opts.update(shell_utils.extract_filters(args.filters))
+    if cs.api_version >= api_versions.APIVersion("3.33"):
+        pools = cs.volumes.get_pools(args.detail, search_opts)
+    else:
+        pools = cs.volumes.get_pools(args.detail)
+    infos = dict()
+    infos.update(pools._info)
+
+    for info in infos['pools']:
+        backend = dict()
+        backend['name'] = info['name']
+        if args.detail:
+            backend.update(info['capabilities'])
+        utils.print_dict(backend)
+
+
+RESET_STATE_RESOURCES = {'volume': utils.find_volume,
+                         'backup': shell_utils.find_backup,
+                         'snapshot': shell_utils.find_volume_snapshot,
+                         'group': shell_utils.find_group,
+                         'group-snapshot': shell_utils.find_group_snapshot}
+
 
 @utils.arg('--group_id',
            metavar='<group_id>',
            default=None,
-           help='Filters results by a group_id. Default=None.',
+           help="Filters results by a group_id. Default=None."
+                "%s" % FILTER_DEPRECATED,
            start_version='3.10')
 @utils.arg('--all-tenants',
            dest='all_tenants',
@@ -52,38 +197,43 @@ from cinderclient.v2.shell import *  # flake8: noqa
 @utils.arg('--name',
            metavar='<name>',
            default=None,
-           help='Filters results by a name. Default=None.')
+           help="Filters results by a name. Default=None. "
+                "%s" % FILTER_DEPRECATED)
 @utils.arg('--display-name',
            help=argparse.SUPPRESS)
 @utils.arg('--status',
            metavar='<status>',
            default=None,
-           help='Filters results by a status. Default=None.')
+           help="Filters results by a status. Default=None. "
+                "%s" % FILTER_DEPRECATED)
 @utils.arg('--bootable',
            metavar='<True|true|False|false>',
            const=True,
            nargs='?',
            choices=['True', 'true', 'False', 'false'],
-           help='Filters results by bootable status. Default=None.')
+           help="Filters results by bootable status. Default=None. "
+                "%s" % FILTER_DEPRECATED)
 @utils.arg('--migration_status',
            metavar='<migration_status>',
            default=None,
-           help='Filters results by a migration status. Default=None. '
-                'Admin only.')
+           help="Filters results by a migration status. Default=None. "
+                "Admin only. "
+                "%s" % FILTER_DEPRECATED)
 @utils.arg('--metadata',
-           type=str,
            nargs='*',
            metavar='<key=value>',
            default=None,
-           help='Filters results by a metadata key and value pair. Require '
-                'volume api version >=3.4. Default=None.')
+           help="Filters results by a metadata key and value pair. "
+                "Default=None. "
+                "%s" % FILTER_DEPRECATED)
 @utils.arg('--image_metadata',
-           type=str,
            nargs='*',
            metavar='<key=value>',
            default=None,
-           help='Filters results by a image metadata key and value pair. '
-                'Default=None.')
+           start_version='3.4',
+           help="Filters results by a image metadata key and value pair. "
+                "Require volume api version >=3.4. Default=None."
+                "%s" % FILTER_DEPRECATED)
 @utils.arg('--marker',
            metavar='<marker>',
            default=None,
@@ -122,8 +272,18 @@ from cinderclient.v2.shell import *  # flake8: noqa
            nargs='?',
            metavar='<tenant>',
            help='Display information from single tenant (Admin only).')
+@utils.arg('--filters',
+           type=six.text_type,
+           nargs='*',
+           start_version='3.33',
+           metavar='<key=value>',
+           default=None,
+           help="Filter key and value pairs. Please use 'cinder list-filters' "
+                "to check enabled filters from server. Use 'key~=value' "
+                "for inexact filtering if the key supports. Default=None.")
 def do_list(cs, args):
     """Lists all volumes."""
+    # pylint: disable=function-redefined
     # NOTE(thingee): Backwards-compatibility with v1 args
     if args.display_name is not None:
         args.name = args.display_name
@@ -141,9 +301,12 @@ def do_list(cs, args):
         if args.metadata else None,
         'glance_metadata': shell_utils.extract_metadata(args,
                                                         type='image_metadata')
-        if args.image_metadata else None,
+        if hasattr(args, 'image_metadata') and args.image_metadata else None,
         'group_id': getattr(args, 'group_id', None),
     }
+    # Update search option with `filters`
+    if hasattr(args, 'filters') and args.filters is not None:
+        search_opts.update(shell_utils.extract_filters(args.filters))
 
     # If unavailable/non-existent fields are specified, these fields will
     # be removed from key_list at the print_list() during key validation.
@@ -170,7 +333,11 @@ def do_list(cs, args):
         setattr(vol, 'attached_to', ','.join(map(str, servers)))
 
     if field_titles:
-        key_list = ['ID'] + field_titles
+        # Remove duplicate fields
+        key_list = ['ID']
+        unique_titles = [k for k in collections.OrderedDict.fromkeys(
+            [x.title().strip() for x in field_titles]) if k != 'Id']
+        key_list.extend(unique_titles)
     else:
         key_list = ['ID', 'Status', 'Name', 'Size', 'Volume Type',
                     'Bootable', 'Attached to']
@@ -187,6 +354,64 @@ def do_list(cs, args):
                      sortby_index=sortby_index)
 
 
+@utils.arg('entity', metavar='<entity>', nargs='+',
+           help='Name or ID of entity to update.')
+@utils.arg('--type', metavar='<type>', default='volume',
+           choices=RESET_STATE_RESOURCES.keys(),
+           help="Type of entity to update. Available resources "
+                "are: 'volume', 'snapshot', 'backup', "
+                "'group' (since 3.20) and "
+                "'group-snapshot' (since 3.19), Default=volume.")
+@utils.arg('--state', metavar='<state>', default=None,
+           help=("The state to assign to the entity. "
+                 "NOTE: This command simply changes the state of the "
+                 "entity in the database with no regard to actual status, "
+                 "exercise caution when using. Default=None, that means the "
+                 "state is unchanged."))
+@utils.arg('--attach-status', metavar='<attach-status>', default=None,
+           help=('This is only used for a volume entity. The attach status '
+                 'to assign to the volume in the database, with no regard '
+                 'to the actual status. Valid values are "attached" and '
+                 '"detached". Default=None, that means the status '
+                 'is unchanged.'))
+@utils.arg('--reset-migration-status',
+           action='store_true',
+           help=('This is only used for a volume entity. Clears the migration '
+                 'status of the volume in the DataBase that indicates the '
+                 'volume is source or destination of volume migration, '
+                 'with no regard to the actual status.'))
+def do_reset_state(cs, args):
+    """Explicitly updates the entity state in the Cinder database.
+
+    Being a database change only, this has no impact on the true state of the
+    entity and may not match the actual state. This can render a entity
+    unusable in the case of changing to the 'available' state.
+    """
+    # pylint: disable=function-redefined
+    failure_count = 0
+    single = (len(args.entity) == 1)
+
+    migration_status = 'none' if args.reset_migration_status else None
+    collector = RESET_STATE_RESOURCES[args.type]
+    argument = (args.state,)
+    if args.type == 'volume':
+        argument += (args.attach_status, migration_status)
+
+    for entity in args.entity:
+        try:
+            collector(cs, entity).reset_state(*argument)
+        except Exception as e:
+            print(e)
+            failure_count += 1
+            msg = "Reset state for entity %s failed: %s" % (entity, e)
+            if not single:
+                print(msg)
+
+    if failure_count == len(args.entity):
+        msg = "Unable to reset the state for the specified entity(s)."
+        raise exceptions.CommandError(msg)
+
+
 @utils.arg('size',
            metavar='<size>',
            nargs='?',
@@ -262,7 +487,6 @@ def do_list(cs, args):
 @utils.arg('--availability_zone',
            help=argparse.SUPPRESS)
 @utils.arg('--metadata',
-           type=str,
            nargs='*',
            metavar='<key=value>',
            default=None,
@@ -462,7 +686,7 @@ def do_group_type_delete(cs, args):
             gtype = shell_utils.find_group_type(cs, group_type)
             cs.group_types.delete(gtype)
             print("Request to delete group type %s has been accepted."
-                  % (group_type))
+                  % group_type)
         except Exception as e:
             failure_count += 1
             print("Delete for group type %s failed: %s" % (group_type, e))
@@ -519,10 +743,6 @@ def do_group_type_key(cs, args):
            metavar='<backup_gigabytes>',
            type=int, default=None,
            help='The new "backup_gigabytes" quota value. Default=None.')
-@utils.arg('--consistencygroups',
-           metavar='<consistencygroups>',
-           type=int, default=None,
-           help='The new "consistencygroups" quota value. Default=None.')
 @utils.arg('--groups',
            metavar='<groups>',
            type=int, default=None,
@@ -574,9 +794,9 @@ def do_quota_update(cs, args):
 @utils.arg('--image_name',
            help=argparse.SUPPRESS)
 @utils.arg('--visibility',
-           metavar='<public|private>',
-           help='Set image visibility to either public or private. '
-                'Default=private.',
+           metavar='<public|private|community|shared>',
+           help='Set image visibility to public, private, community or '
+                'shared. Default=private.',
            default='private',
            start_version='3.1')
 @utils.arg('--protected',
@@ -625,6 +845,7 @@ def do_backup_update(cs, args):
         raise exceptions.ClientException(code=1, message=msg)
 
     shell_utils.find_backup(cs, args.backup).update(**kwargs)
+    print("Request to update backup '%s' has been accepted." % args.backup)
 
 
 @api_versions.wraps('3.7')
@@ -710,7 +931,8 @@ def do_cluster_disable(cs, args):
            metavar='<marker>',
            default=None,
            help='Begin returning volumes that appear later in the volume '
-                'list than that represented by this volume id. '
+                'list than that represented by this reference. This reference '
+                'should be json like. '
                 'Default=None.')
 @utils.arg('--limit',
            metavar='<limit>',
@@ -726,9 +948,11 @@ def do_cluster_disable(cs, args):
            help=(('Comma-separated list of sort keys and directions in the '
                   'form of <key>[:<asc|desc>]. '
                   'Valid keys: %s. '
-                  'Default=None.') % ', '.join(base.SORT_KEY_VALUES)))
+                  'Default=None.'
+                  ) % ', '.join(base.SORT_MANAGEABLE_KEY_VALUES)))
 def do_manageable_list(cs, args):
     """Lists all manageable volumes."""
+    # pylint: disable=function-redefined
     detailed = strutils.bool_from_string(args.detailed)
     volumes = cs.volumes.list_manageable(host=args.host, detailed=detailed,
                                          marker=args.marker, limit=args.limit,
@@ -746,16 +970,58 @@ def do_manageable_list(cs, args):
            nargs='?',
            type=int,
            const=1,
-           default=0,
+           default=utils.env('ALL_TENANTS', default=0),
            help='Shows details for all tenants. Admin only.')
+@utils.arg('--filters',
+           type=six.text_type,
+           nargs='*',
+           start_version='3.33',
+           metavar='<key=value>',
+           default=None,
+           help="Filter key and value pairs. Please use 'cinder list-filters' "
+                "to check enabled filters from server. Use 'key~=value' "
+                "for inexact filtering if the key supports. Default=None.")
 def do_group_list(cs, args):
     """Lists all groups."""
-    groups = cs.groups.list()
+    search_opts = {'all_tenants': args.all_tenants}
+
+    # Update search option with `filters`
+    if hasattr(args, 'filters') and args.filters is not None:
+        search_opts.update(shell_utils.extract_filters(args.filters))
+
+    groups = cs.groups.list(search_opts=search_opts)
 
     columns = ['ID', 'Status', 'Name']
     utils.print_list(groups, columns)
 
 
+@api_versions.wraps('3.13')
+@utils.arg('--list-volume',
+           dest='list_volume',
+           metavar='<False|True>',
+           nargs='?',
+           type=bool,
+           const=True,
+           default=False,
+           help='Shows volumes included in the group.',
+           start_version='3.25')
+@utils.arg('group',
+           metavar='<group>',
+           help='Name or ID of a group.')
+def do_group_show(cs, args):
+    """Shows details of a group."""
+    info = dict()
+    if getattr(args, 'list_volume', None):
+        group = shell_utils.find_group(cs, args.group,
+                                       list_volume=args.list_volume)
+    else:
+        group = shell_utils.find_group(cs, args.group)
+    info.update(group._info)
+
+    info.pop('links', None)
+    utils.print_dict(info)
+
+
 @api_versions.wraps('3.13')
 @utils.arg('grouptype',
            metavar='<group-type>',
@@ -901,9 +1167,77 @@ def do_group_update(cs, args):
         raise exceptions.ClientException(code=1, message=msg)
 
     shell_utils.find_group(cs, args.group).update(**kwargs)
+    print("Request to update group '%s' has been accepted." % args.group)
+
+
+@api_versions.wraps('3.38')
+@utils.arg('group',
+           metavar='<group>',
+           help='Name or ID of the group.')
+def do_group_enable_replication(cs, args):
+    """Enables replication for group."""
+
+    shell_utils.find_group(cs, args.group).enable_replication()
+
+
+@api_versions.wraps('3.38')
+@utils.arg('group',
+           metavar='<group>',
+           help='Name or ID of the group.')
+def do_group_disable_replication(cs, args):
+    """Disables replication for group."""
+
+    shell_utils.find_group(cs, args.group).disable_replication()
+
+
+@api_versions.wraps('3.38')
+@utils.arg('group',
+           metavar='<group>',
+           help='Name or ID of the group.')
+@utils.arg('--allow-attached-volume',
+           action='store_true',
+           default=False,
+           help='Allows or disallows group with '
+                'attached volumes to be failed over.')
+@utils.arg('--secondary-backend-id',
+           metavar='<secondary_backend_id>',
+           help='Secondary backend id. Default=None.')
+def do_group_failover_replication(cs, args):
+    """Fails over replication for group."""
+
+    shell_utils.find_group(cs, args.group).failover_replication(
+        allow_attached_volume=args.allow_attached_volume,
+        secondary_backend_id=args.secondary_backend_id)
+
+
+@api_versions.wraps('3.38')
+@utils.arg('group',
+           metavar='<group>',
+           help='Name or ID of the group.')
+def do_group_list_replication_targets(cs, args):
+    """Lists replication targets for group.
+
+    Example value for replication_targets:
+
+    .. code-block: json
+
+        {
+            'replication_targets': [{'backend_id': 'vendor-id-1',
+                                     'unique_key': 'val1',
+                                     ......},
+                                    {'backend_id': 'vendor-id-2',
+                                     'unique_key': 'val2',
+                                     ......}]
+        }
+    """
+
+    rc, replication_targets = shell_utils.find_group(
+        cs, args.group).list_replication_targets()
+    rep_targets = replication_targets.get('replication_targets')
+    if rep_targets and len(rep_targets) > 0:
+        utils.print_list(rep_targets, [key for key in rep_targets[0].keys()])
 
 
-@api_versions.wraps('3.14')
 @utils.arg('--all-tenants',
            dest='all_tenants',
            metavar='<0|1>',
@@ -915,11 +1249,22 @@ def do_group_update(cs, args):
 @utils.arg('--status',
            metavar='<status>',
            default=None,
-           help='Filters results by a status. Default=None.')
+           help="Filters results by a status. Default=None. "
+                "%s" % FILTER_DEPRECATED)
 @utils.arg('--group-id',
            metavar='<group_id>',
            default=None,
-           help='Filters results by a group ID. Default=None.')
+           help="Filters results by a group ID. Default=None. "
+                "%s" % FILTER_DEPRECATED)
+@utils.arg('--filters',
+           type=six.text_type,
+           nargs='*',
+           start_version='3.33',
+           metavar='<key=value>',
+           default=None,
+           help="Filter key and value pairs. Please use 'cinder list-filters' "
+                "to check enabled filters from server. Use 'key~=value' "
+                "for inexact filtering if the key supports. Default=None.")
 def do_group_snapshot_list(cs, args):
     """Lists all group snapshots."""
 
@@ -930,6 +1275,9 @@ def do_group_snapshot_list(cs, args):
         'status': args.status,
         'group_id': args.group_id,
     }
+    # Update search option with `filters`
+    if hasattr(args, 'filters') and args.filters is not None:
+        search_opts.update(shell_utils.extract_filters(args.filters))
 
     group_snapshots = cs.group_snapshots.list(search_opts=search_opts)
 
@@ -998,7 +1346,7 @@ def do_group_snapshot_delete(cs, args):
                                       "group snapshots.")
 
 
-@api_versions.wraps('3.7')
+@api_versions.wraps('3.0')
 @utils.arg('--host', metavar='<hostname>', default=None,
            help='Host name. Default=None.')
 @utils.arg('--binary', metavar='<binary>', default=None,
@@ -1008,12 +1356,17 @@ def do_group_snapshot_delete(cs, args):
            const=True,
            nargs='?',
            default=False,
+           start_version='3.7',
            help='Enables or disables display of '
                 'Replication info for c-vol services. Default=False.')
 def do_service_list(cs, args):
     """Lists all services. Filter by host and service binary."""
-    replication = strutils.bool_from_string(args.withreplication,
-                                            strict=True)
+    if hasattr(args, 'withreplication'):
+        replication = strutils.bool_from_string(args.withreplication,
+                                                strict=True)
+    else:
+        replication = False
+
     result = cs.services.list(host=args.host, binary=args.binary)
     columns = ["Binary", "Host", "Zone", "Status", "State", "Updated_at"]
     if cs.api_version.matches('3.7'):
@@ -1040,7 +1393,8 @@ def do_service_list(cs, args):
            metavar='<marker>',
            default=None,
            help='Begin returning volumes that appear later in the volume '
-                'list than that represented by this volume id. '
+                'list than that represented by this reference. This reference '
+                'should be json like. '
                 'Default=None.')
 @utils.arg('--limit',
            metavar='<limit>',
@@ -1056,9 +1410,11 @@ def do_service_list(cs, args):
            help=(('Comma-separated list of sort keys and directions in the '
                   'form of <key>[:<asc|desc>]. '
                   'Valid keys: %s. '
-                  'Default=None.') % ', '.join(base.SORT_KEY_VALUES)))
+                  'Default=None.'
+                  ) % ', '.join(base.SORT_MANAGEABLE_KEY_VALUES)))
 def do_snapshot_manageable_list(cs, args):
     """Lists all manageable snapshots."""
+    # pylint: disable=function-redefined
     detailed = strutils.bool_from_string(args.detailed)
     snapshots = cs.volume_snapshots.list_manageable(host=args.host,
                                                     detailed=detailed,
@@ -1080,6 +1436,19 @@ def do_api_version(cs, args):
     utils.print_list(response, columns)
 
 
+@api_versions.wraps("3.40")
+@utils.arg(
+    'snapshot',
+    metavar='<snapshot>',
+    help='Name or ID of the snapshot to restore. The snapshot must be the '
+         'most recent one known to cinder.')
+def do_revert_to_snapshot(cs, args):
+    """Revert a volume to the specified snapshot."""
+    snapshot = shell_utils.find_volume_snapshot(cs, args.snapshot)
+    volume = utils.find_volume(cs, snapshot.volume_id)
+    volume.revert_to_snapshot(snapshot)
+
+
 @api_versions.wraps("3.3")
 @utils.arg('--marker',
            metavar='<marker>',
@@ -1104,23 +1473,37 @@ def do_api_version(cs, args):
 @utils.arg('--resource_uuid',
            metavar='<resource_uuid>',
            default=None,
-           help='Filters results by a resource uuid. Default=None.')
+           help="Filters results by a resource uuid. Default=None. "
+                "%s" % FILTER_DEPRECATED)
 @utils.arg('--resource_type',
            metavar='<type>',
            default=None,
-           help='Filters results by a resource type. Default=None.')
+           help="Filters results by a resource type. Default=None. "
+                "%s" % FILTER_DEPRECATED)
 @utils.arg('--event_id',
            metavar='<id>',
            default=None,
-           help='Filters results by event id. Default=None.')
+           help="Filters results by event id. Default=None. "
+                "%s" % FILTER_DEPRECATED)
 @utils.arg('--request_id',
            metavar='<request_id>',
            default=None,
-           help='Filters results by request id. Default=None.')
+           help="Filters results by request id. Default=None. "
+                "%s" % FILTER_DEPRECATED)
 @utils.arg('--level',
            metavar='<level>',
            default=None,
-           help='Filters results by the message level. Default=None.')
+           help="Filters results by the message level. Default=None. "
+                "%s" % FILTER_DEPRECATED)
+@utils.arg('--filters',
+           type=six.text_type,
+           nargs='*',
+           start_version='3.33',
+           metavar='<key=value>',
+           default=None,
+           help="Filter key and value pairs. Please use 'cinder list-filters' "
+                "to check enabled filters from server. Use 'key~=value' "
+                "for inexact filtering if the key supports. Default=None.")
 def do_message_list(cs, args):
     """Lists all messages."""
     search_opts = {
@@ -1128,6 +1511,9 @@ def do_message_list(cs, args):
         'event_id': args.event_id,
         'request_id': args.request_id,
     }
+    # Update search option with `filters`
+    if hasattr(args, 'filters') and args.filters is not None:
+        search_opts.update(shell_utils.extract_filters(args.filters))
     if args.resource_type:
         search_opts['resource_type'] = args.resource_type.upper()
     if args.level:
@@ -1181,6 +1567,7 @@ def do_message_delete(cs, args):
         raise exceptions.CommandError("Unable to delete any of the specified "
                                       "messages.")
 
+
 @utils.arg('--all-tenants',
            dest='all_tenants',
            metavar='<0|1>',
@@ -1197,7 +1584,8 @@ def do_message_delete(cs, args):
 @utils.arg('--name',
            metavar='<name>',
            default=None,
-           help='Filters results by a name. Default=None.')
+           help="Filters results by a name. Default=None. "
+                "%s" % FILTER_DEPRECATED)
 @utils.arg('--display-name',
            help=argparse.SUPPRESS)
 @utils.arg('--display_name',
@@ -1205,11 +1593,13 @@ def do_message_delete(cs, args):
 @utils.arg('--status',
            metavar='<status>',
            default=None,
-           help='Filters results by a status. Default=None.')
+           help="Filters results by a status. Default=None. "
+                "%s" % FILTER_DEPRECATED)
 @utils.arg('--volume-id',
            metavar='<volume-id>',
            default=None,
-           help='Filters results by a volume ID. Default=None.')
+           help="Filters results by a volume ID. Default=None. "
+                "%s" % FILTER_DEPRECATED)
 @utils.arg('--volume_id',
            help=argparse.SUPPRESS)
 @utils.arg('--marker',
@@ -1240,10 +1630,21 @@ def do_message_delete(cs, args):
            metavar='<key=value>',
            default=None,
            start_version='3.22',
-           help='Filters results by a metadata key and value pair. Require '
-                'volume api version >=3.22. Default=None.')
+           help="Filters results by a metadata key and value pair. Require "
+                "volume api version >=3.22. Default=None. "
+                "%s" % FILTER_DEPRECATED)
+@utils.arg('--filters',
+           type=six.text_type,
+           nargs='*',
+           start_version='3.33',
+           metavar='<key=value>',
+           default=None,
+           help="Filter key and value pairs. Please use 'cinder list-filters' "
+                "to check enabled filters from server. Use 'key~=value' "
+                "for inexact filtering if the key supports. Default=None.")
 def do_snapshot_list(cs, args):
     """Lists all snapshots."""
+    # pylint: disable=function-redefined
     all_tenants = (1 if args.tenant else
                    int(os.environ.get("ALL_TENANTS", args.all_tenants)))
 
@@ -1266,6 +1667,10 @@ def do_snapshot_list(cs, args):
         'metadata': metadata
     }
 
+    # Update search option with `filters`
+    if hasattr(args, 'filters') and args.filters is not None:
+        search_opts.update(shell_utils.extract_filters(args.filters))
+
     snapshots = cs.volume_snapshots.list(search_opts=search_opts,
                                          marker=args.marker,
                                          limit=args.limit,
@@ -1276,6 +1681,7 @@ def do_snapshot_list(cs, args):
                      ['ID', 'Volume ID', 'Status', 'Name', 'Size'],
                      sortby_index=sortby_index)
 
+
 @api_versions.wraps('3.27')
 @utils.arg('--all-tenants',
            dest='all_tenants',
@@ -1283,16 +1689,18 @@ def do_snapshot_list(cs, args):
            nargs='?',
            type=int,
            const=1,
-           default=0,
+           default=utils.env('ALL_TENANTS', default=0),
            help='Shows details for all tenants. Admin only.')
 @utils.arg('--volume-id',
            metavar='<volume-id>',
            default=None,
-           help='Filters results by a volume ID. Default=None.')
+           help="Filters results by a volume ID. Default=None. "
+                "%s" % FILTER_DEPRECATED)
 @utils.arg('--status',
            metavar='<status>',
            default=None,
-           help='Filters results by a status. Default=None.')
+           help="Filters results by a status. Default=None. "
+                "%s" % FILTER_DEPRECATED)
 @utils.arg('--marker',
            metavar='<marker>',
            default=None,
@@ -1302,7 +1710,7 @@ def do_snapshot_list(cs, args):
 @utils.arg('--limit',
            metavar='<limit>',
            default=None,
-           help='Maximum number of attachemnts to return. Default=None.')
+           help='Maximum number of attachments to return. Default=None.')
 @utils.arg('--sort',
            metavar='<key>[:<direction>]',
            default=None,
@@ -1316,19 +1724,32 @@ def do_snapshot_list(cs, args):
            nargs='?',
            metavar='<tenant>',
            help='Display information from single tenant (Admin only).')
+@utils.arg('--filters',
+           type=six.text_type,
+           nargs='*',
+           start_version='3.33',
+           metavar='<key=value>',
+           default=None,
+           help="Filter key and value pairs. Please use 'cinder list-filters' "
+                "to check enabled filters from server. Use 'key~=value' "
+                "for inexact filtering if the key supports. Default=None.")
 def do_attachment_list(cs, args):
     """Lists all attachments."""
     search_opts = {
-        'all_tenants': args.all_tenants,
+        'all_tenants': 1 if args.tenant else args.all_tenants,
+        'project_id': args.tenant,
         'status': args.status,
         'volume_id': args.volume_id,
     }
+    # Update search option with `filters`
+    if hasattr(args, 'filters') and args.filters is not None:
+        search_opts.update(shell_utils.extract_filters(args.filters))
 
     attachments = cs.attachments.list(search_opts=search_opts,
                                       marker=args.marker,
                                       limit=args.limit,
                                       sort=args.sort)
-    columns = ['ID', 'Volume ID', 'Status', 'Instance']
+    columns = ['ID', 'Volume ID', 'Status', 'Server ID']
     if args.sort:
         sortby_index = None
     else:
@@ -1358,10 +1779,9 @@ def do_attachment_show(cs, args):
 @utils.arg('volume',
            metavar='<volume>',
            help='Name or ID of volume or volumes to attach.')
-@utils.arg('--instance',
-           metavar='<instance>',
-           default=None,
-           help='UUID of Instance attaching to. Default=None.')
+@utils.arg('server_id',
+           metavar='<server_id>',
+           help='ID of server attaching to.')
 @utils.arg('--connect',
            metavar='<connect>',
            default=False,
@@ -1390,7 +1810,7 @@ def do_attachment_show(cs, args):
 @utils.arg('--multipath',
            metavar='<multipath>',
            default=False,
-           help='OS type. Default=False.')
+           help='Use multipath. Default=False.')
 @utils.arg('--mountpoint',
            metavar='<mountpoint>',
            default=None,
@@ -1406,10 +1826,12 @@ def do_attachment_create(cs, args):
                      'platform': args.platform,
                      'host': args.host,
                      'os_type': args.ostype,
-                     'multipath': args.multipath}
-    attachment = cs.attachments.create(args.volume,
+                     'multipath': args.multipath,
+                     'mountpoint': args.mountpoint}
+    volume = utils.find_volume(cs, args.volume)
+    attachment = cs.attachments.create(volume.id,
                                        connector,
-                                       args.instance)
+                                       args.server_id)
     connector_dict = attachment.pop('connection_info', None)
     utils.print_dict(attachment)
     if connector_dict:
@@ -1443,7 +1865,7 @@ def do_attachment_create(cs, args):
 @utils.arg('--multipath',
            metavar='<multipath>',
            default=False,
-           help='OS type. Default=False.')
+           help='Use multipath. Default=False.')
 @utils.arg('--mountpoint',
            metavar='<mountpoint>',
            default=None,
@@ -1459,7 +1881,8 @@ def do_attachment_update(cs, args):
                  'platform': args.platform,
                  'host': args.host,
                  'os_type': args.ostype,
-                 'multipath': args.multipath}
+                 'multipath': args.multipath,
+                 'mountpoint': args.mountpoint}
     attachment = cs.attachments.update(args.attachment,
                                        connector)
     attachment_dict = attachment.to_dict()
@@ -1477,3 +1900,63 @@ def do_attachment_delete(cs, args):
     """Delete an attachment for a cinder volume."""
     for attachment in args.attachment:
         cs.attachments.delete(attachment)
+
+
+@api_versions.wraps('3.0')
+def do_version_list(cs, args):
+    """List all API versions."""
+    result = cs.services.server_api_version()
+    if 'min_version' in dir(result[0]):
+        columns = ["Id", "Status", "Updated", "Min Version", "Version"]
+    else:
+        columns = ["Id", "Status", "Updated"]
+
+    print("Client supported API versions:")
+    print("Minimum version %(v)s" %
+          {'v': api_versions.MIN_VERSION})
+    print("Maximum version %(v)s" %
+          {'v': api_versions.MAX_VERSION})
+
+    print("\nServer supported API versions:")
+    utils.print_list(result, columns)
+
+
+@api_versions.wraps('3.32')
+@utils.arg('level',
+           metavar='<log-level>',
+           choices=('INFO', 'WARNING', 'ERROR', 'DEBUG',
+                    'info', 'warning', 'error', 'debug'),
+           help='Desired log level.')
+@utils.arg('--binary',
+           choices=('', '*', 'cinder-api', 'cinder-volume', 'cinder-scheduler',
+                    'cinder-backup'),
+           default='',
+           help='Binary to change.')
+@utils.arg('--server',
+           default='',
+           help='Host or cluster value for service.')
+@utils.arg('--prefix',
+           default='',
+           help='Prefix for the log. ie: "cinder.volume.drivers.".')
+def do_service_set_log(cs, args):
+    cs.services.set_log_levels(args.level, args.binary, args.server,
+                               args.prefix)
+
+
+@api_versions.wraps('3.32')
+@utils.arg('--binary',
+           choices=('', '*', 'cinder-api', 'cinder-volume', 'cinder-scheduler',
+                    'cinder-backup'),
+           default='',
+           help='Binary to query.')
+@utils.arg('--server',
+           default='',
+           help='Host or cluster value for service.')
+@utils.arg('--prefix',
+           default='',
+           help='Prefix for the log. ie: "sqlalchemy.".')
+def do_service_get_log(cs, args):
+    log_levels = cs.services.get_log_levels(args.binary, args.server,
+                                            args.prefix)
+    columns = ('Binary', 'Host', 'Prefix', 'Level')
+    utils.print_list(log_levels, columns)
